#include <iostream>
#include "gpuTimer.cuh"
#include "timer.hxx"

#include <thrust/device_vector.h>
#include <thrust/transform.h>
#include <thrust/functional.h>

void saxpy1(double A, 
            const std::vector<double> &X, 
            std::vector<double> &Y)
{
    size_t n = X.size(), i;
    for (i=0; i<n; i++)
      Y[i] += A*X[i];
}

void saxpy2(double A, 
            const std::vector<double> &X, 
            std::vector<double> &Y)
{
    size_t n = X.size(), i;
    #pragma omp parallel for
    for (i=0; i<n; i++)
      Y[i] += A*X[i];
}

struct saxpy_functor {

  const double _a;
  saxpy_functor(double a) : _a(a) {}

  __host__ __device__ 
  double operator() (double x, double y) const
  {
    return _a*x + y;
  }
};

template<typename T>
void saxpy3(const std::vector<double> &X, 
             std::vector<double> &Y,
             T & f)
{
  std::transform(X.begin(), X.end(),  
                 Y.begin(), Y.begin(),
                 f
                );        
}

template<typename T>
void saxpy4(const thrust::device_vector<double> &X, 
            thrust::device_vector<double> &Y,
            T & f)
{
  thrust::transform(X.begin(), X.end(),  
                    Y.begin(), Y.begin(),
                    f
                   );
  cudaDeviceSynchronize();       
}

int main(int argc, char ** argv)
{
  size_t n = argc > 1 ? atoll(argv[1]) : 100000000L;

  std::vector<double> h_X(n, 2.0);
  std::vector<double> h_Y(n, 1.0);
  thrust::device_vector<double> d_X(n, 2.0);
  thrust::device_vector<double> d_Y(n, 1.0);

  double a = 1.5;
// ------------------------------------------
  auto f = [a] (double u, double v) { return a * u + v; };
  auto g = [a] __device__ (double u, double v) { return a * u + v; };
  auto h = [a] __host__ __device__ (double u, double v) { return a * u + v; };
// ------------------------------------------

  Timer T_cpu1, T_cpu2, T_cpu3, T_cpu4 ;
  CudaTimer T_gpu1, T_gpu2, T_gpu3;

  T_cpu1.start();
  saxpy1(a, h_X, h_Y);
  T_cpu1.stop();
  std::cout << "saxpy CPU :               time " << T_cpu1.elapsed() << " s" << std::endl; 

  T_cpu2.start();
  saxpy2(a, h_X, h_Y);
  T_cpu2.stop();
  std::cout << "saxpy CPU (openmp) :      time " << T_cpu2.elapsed() << " s" << std::endl;

  saxpy_functor ff(a);

  T_cpu3.start();
  saxpy3(h_X, h_Y, ff);
  T_cpu3.stop();
  std::cout << "saxpy CPU functor class : time " << T_cpu3.elapsed() << " s" << std::endl;

  T_gpu1.start();
  saxpy4(d_X, d_Y, ff);
  T_gpu1.stop();
  std::cout << "saxpy GPU functor class : time " << T_gpu1.elapsed() << " s" << std::endl;

  T_cpu4.start();
  saxpy3(h_X, h_Y, f);
  T_cpu4.stop();
  std::cout << "saxpy CPU lambda :        time " << T_cpu4.elapsed() << " s" << std::endl; 

  T_cpu4.start();
  saxpy3(h_X, h_Y, h);
  T_cpu4.stop();
  std::cout << "saxpy CPU lambda :        time " << T_cpu4.elapsed() << " s" << std::endl; 

  T_gpu2.start();
  saxpy4(d_X, d_Y, g);
  T_gpu2.stop();
  std::cout << "saxpy GPU lambda :        time " << T_gpu2.elapsed() << " s" << std::endl; 

  T_gpu3.start();
  saxpy4(d_X, d_Y, h);
  T_gpu3.stop();
  std::cout << "saxpy GPU lambda :        time " << T_gpu3.elapsed() << " s" << std::endl; 

  return 0;
}
